function A_star(coordinates, mode, LVals)
%{
******************************************************************************************
*                                                                                        *
*    A* search with homotopy class constraints                                           *
*    Copyright (C) 2010  Subhrajit Bhattacharya                                          *
*                                                                                        *
*    This program is free software: you can redistribute it and/or modify                *
*    it under the terms of the GNU General Public License as published by                *
*    the Free Software Foundation, either version 3 of the License, or                   *
*    (at your option) any later version.                                                 *
*                                                                                        *
*    This program is distributed in the hope that it will be useful,                     *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of                      *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                       *
*    GNU General Public License for more details <http://www.gnu.org/licenses/>.         *
*                                                                                        *
*                                                                                        *
*    Contact: subhrajit@gmail.com, http://fling.seas.upenn.edu/~subhrabh/                *
*                                                                                        *
*                                                                                        *
******************************************************************************************
%}

clear global
global Map2D N M z0 zs BlockedLVals Obs
global g_score h_score f_score i_score came_from

showExpansions = false;

if nargin < 2
    mode = 'b';
    LVals = [];
end
global ComputedTransitions
start = coordinates(1,:);
stop = coordinates(2,:);
SkipSearch = false;

ComputedTransitions = [];
zs = start(1) + i*start(2);
Obs = [];
z0 = [];

DefineEnv;


% fh = @(x,y) min([abs(f(x+i*y)) 1e3]);
% ezmeshc(fh, [0, M, 0, N] );
% pause;

% Blocked L-values (homotopy class)
switch mode
    case 'b'    % List of blocked homotopy class provided
        BlockedLVals = LVals;
        TargetLVal = [];
    case 'c'    % Constraint homotopy class
        BlockedLVals = [];
        TargetLVal = LVals(1);
    case 'k'
        SkipSearch = true;
        ExpansionCount = 0;
        PathNodes = coordinates(1,:);
        for a=1:(size(coordinates,1)-1)
            dX = coordinates(a+1,1)-coordinates(a,1);
            dY = coordinates(a+1,2)-coordinates(a,2);
            segLen = max(abs([dX, dY]));
            PathNodes((end+1):(end+segLen),:) = [coordinates(a,1)+round((1:segLen)'*dX/segLen), coordinates(a,2)+round((1:segLen)'*dY/segLen)];
        end
end

% Generate Obstacle map
Map2D = zeros(N,M);
for a=1:size(Obs,1)
    Map2D(Obs(a,1),Obs(a,2)) = 1;
end

% Perform the search
if ~SkipSearch
    [StopIndex, ExpansionCount] = search(start, stop, TargetLVal);
    PathNodes = reconstruct_path(StopIndex, start);
    PathNodes = PathNodes(end:-1:1,:);
end

disp(sprintf('\nThe path coords:'));
disp(PathNodes);
thisPathLval = CplxI(PathNodes);
disp(sprintf('Number of states expanded: %d', ExpansionCount));
if (imag(thisPathLval) >= 0)
    disp(sprintf('The L-val of this path = %f+%fi', real(thisPathLval), imag(thisPathLval)));
else
    disp(sprintf('The L-val of this path = %f%fi \n', real(thisPathLval), imag(thisPathLval)));
end


% Plot results
figH = figure(101);
axis equal;
axis([0 N 0 M]);
hold on;

for a=1:N
    for b=1:M
        if Map2D(a,b)
            plotCell(a,b,[0.6 0.7 0.6]);
        elseif any(PathNodes(find(a==PathNodes(:,1)),2)==b)
            plotCell(a,b,[1 0 0]);
        else
            %plotCell(a,b,[1 1 1]);
        end
        %thisInd = X2I([a,b]);
        %text(a,b,sprintf('(%f,\n%f,\n%f+i%f)',g_score(thisInd),f_score(thisInd),real(i_score(thisInd)),imag(i_score(thisInd))));
    end
end

if showExpansions
    for xind = 1:length(came_from)
        if came_from(xind)==0
            continue;
        end
        thisX = I2X(xind);
        fromX = I2X(came_from(xind));
        %thisLval = real(thisLval)^2 + i*sqrt(imag(thisLval));
        colr = [0 1 1]; %([1 0 1] + [real(thisLval) 0 imag(thisLval)]/abs(thisLval))/2;
        plot([thisX(1),fromX(1)], [thisX(2),fromX(2)], '-', 'Color', colr);
        %if ~any(PathNodes(find(fromX(1)==PathNodes(:,1)),2)==fromX(2))
            plot(fromX(1), fromX(2), 'x', 'Color', colr);
        %end
    end
end

for a=1:length(z0)
    plot3(real(z0(a)),imag(z0(a)), 20, 'bo');
end

text(coordinates(1,1), coordinates(1,2)+1, 20, '\bf\it\fontsize{15}\fontname{Times New Roman}  z_{2s}');
text(coordinates(end,1), coordinates(end,2)+1, 20, '\bf\it\fontsize{15}\fontname{Times New Roman}  z_{2g}');
plot3(coordinates(:,1), coordinates(:,end), 20*ones(size(coordinates,1),1), 'b*');

%pause;

% ------------------------------------------

function plotCell(x,y,c,typ,styl)
patchX = [x-0.5; x+0.5; x+0.5; x-0.5];
patchY = [y+0.5; y+0.5; y-0.5; y-0.5];
patchZ = [10; 10; 10; 10];
if nargin < 4
    typ = 1;
end
switch(typ)
    case 1
        patch(patchX, patchY, patchZ, c, 'EdgeColor', 'none');
    case 2
        patch(patchX, patchY, patchZ, 'w', 'EdgeColor', c);
    case 3
        if nargin < 5
            styl = '-';
        end
        plot(x, y, styl, 'Color', c, 'LineWidth', 2);
end

% ======================================================================

function dI = Delta_i(z1, z2)

if length(z1)>1
    z1 = z1(1)+i*z1(2);
end
if length(z2)>1
    z2 = z2(1)+i*z2(2);
end
% Adaptive step size
%zm = (z1+z2)/2;
%Ofzm = log(abs(f(zm)));
%del = 0.1*0.1^ceil(min([max([1 Ofzm-6]) 2])); %pause;
del = 1e-3; %0.001;
% Integration
Dz = z2 - z1;
delDz = del*Dz;
dI = 0+0*i;
p = (del/2):del:(1-del/2);
    cz = z1 + p*Dz;
    dI = sum(f(cz))*delDz;
%end

function fz = f(z)
global z0 N M zs
fz = ones(1, length(z));
for a=1:length(z0)
    fz = fz .* (z - z0(a));
end
fz = ((((z-(M+i*N)).*z).^(length(z0)/2))) ./ fz;
%fz = (sqrt(M*N)^(length(z0)/2)) ./ fz;

function zI = CplxI(PathNodes)
zI = 0+0*i;
for a = 1:(size(PathNodes,1)-1)
    zI = zI + Delta_i(PathNodes(a,:),PathNodes(a+1,:));
end

% ======================================================================
% ======================================================================

function [StopIndex, ExpansionCount] = search(start, stop, TargetLVal)
global Map2D N M BlockedLVals
global ExistingLVals
global g_score h_score f_score i_score came_from
global openset closedset
V = 1000;

closedset = [];                 % The set of nodes already evaluated.
openset = X2I(start,0); % The set of tentative nodes to be evaluated.

buffSize = 1000;
g_score = inf(1,buffSize);
f_score = inf(1,buffSize);
came_from = zeros(1,buffSize);
%i_score = zeros(1,buffSize) + zeros(1,buffSize)*i;

g_score(X2I(start,0)) = 0;                        % Distance from start along optimal path.
h_score(X2I(start,0)) = heuristic_estimate_of_distance(start, stop);
f_score(X2I(start,0)) = h_score(X2I(start,0));           % Estimated total distance from start to goal through y.
%i_score(X2I(start,0)) = 0+0i;

while length(openset)>0

    xindTmp = find(f_score(openset)==min(f_score(openset)));
    xind = openset(xindTmp(1));

    if (~isempty(TargetLVal) && xind==X2I(stop,TargetLVal))
        StopIndex = xind;
        ExpansionCount = length(closedset);
        return;
    elseif (isempty(TargetLVal) && all(stop==I2X(xind)))
        [x,xL] = I2X(xind);
        % adaptiveThresh = thresh * 10^log(max([1 abs(Lval(a))]));
        if ~any(abs(ExistingLVals{x(1),x(2)}(xL) - BlockedLVals) < 0.01)
            StopIndex = xind;
            ExpansionCount = length(closedset);
            return;
        end
    end

    openset = setdiff(openset,xind);
    closedset(end+1) = xind;
    for yind = neighbor_nodes(xind)
        if yind > buffSize
            buffSize = yind + 1000;
            came_from(end+1:buffSize) = 0;
            g_score(end+1:buffSize) = inf;
            f_score(end+1:buffSize) = inf;
        end

        if any(yind==closedset)
            continue;
        end

        x = I2X(xind);
        y = I2X(yind);
        %tentative_i_score = i_score(xind) + Delta_i(x,y);
        tentative_g_score = g_score(xind) + dist_between(x,y); % + V * abs(i_score(yind) + Delta_i(y,x) + Stop2StartI);

        if ~any(yind==openset)
            openset(end+1) = yind;
            tentative_is_better = true;
        elseif tentative_g_score < g_score(yind)
            tentative_is_better = true;
        else
            tentative_is_better = false;
        end

        if tentative_is_better==true
            came_from(yind) = xind;
            g_score(yind) = tentative_g_score;
            %old_i_score = i_score(yind);
            %i_score(yind) = tentative_i_score; % + Delta_i(x,y);
            h_score(yind) = heuristic_estimate_of_distance(y, stop);
            f_score(yind) = g_score(yind) + h_score(yind);
        end

    end
end

StopIndex = X2I(start,0);
ExpansionCount = length(closedset);

% ======================================================================

function ii = X2I(xx, Lval)
global N M
global ExistingLVals
thresh = 0.1; %0.01;

if isempty(ExistingLVals)
    ExistingLVals = cell(N,M);
end

if isempty(Lval) && size(xx,1)==1
    
    layerNo = 1;
    layerNos = 1:length(ExistingLVals{xx(1,1),xx(1,2)});
    ii = N*M*(layerNos-1) + N*(xx(1,2)-1) + xx(1,1);
    
else
    
    for a=1:size(xx,1)
        adaptiveThresh = thresh; % * (max([1 abs(Lval(a))]));
        ExistFlags = (abs(ExistingLVals{xx(a,1),xx(a,2)} - Lval(a)) < adaptiveThresh);
        if any(ExistFlags)
            layerNo(a) = find(ExistFlags);
            ExistingLVals{xx(a,1),xx(a,2)}(layerNo(a)) = 0.7*ExistingLVals{xx(a,1),xx(a,2)}(layerNo(a)) + 0.3*Lval(a);
        else
            layerNo(a) = length(ExistingLVals{xx(a,1),xx(a,2)}) + 1;
            ExistingLVals{xx(a,1),xx(a,2)}(layerNo(a)) = Lval(a);
        end
    end
    
    ii = N*M*(layerNo-1) + N*(xx(:,2)-1) + xx(:,1);
    
end
        

% ======================================================================

function [xx,LNo,LVal] = I2X(ind)
global N M ExistingLVals
ii = mod(ind-1,N*M) + 1;
xx(:,1) = mod(ii-1, N) + 1;
xx(:,2) = floor((ii-1)/N) + 1;
if nargout>1
    LNo = floor((ind-1)/(N*M)) + 1;
end
if nargout>2
    for a=1:length(ind)
        LVal(a) = ExistingLVals{xx(a,1),xx(a,2)}(LNo(a));
    end
end

% ======================================================================

function h = heuristic_estimate_of_distance(a, b)
dx = a(1) - b(1);
dy = a(2) - b(2);
h = sqrt(dx*dx + dy*dy);

% ======================================================================

function d = dist_between(a, b)
dx = a(1) - b(1);
dy = a(2) - b(2);
d = sqrt(dx*dx + dy*dy);

% ======================================================================

function yi = neighbor_nodes(xi)
global Map2D N M
global ExistingLVals
global ComputedTransitions
% 8-connected grid
[x, xL] = I2X(xi);
yi = [];
neiCount = 0;
for a = -1:1
    for b = -1:1
        if a==0 && b==0
            continue;
        end
        neiCount = neiCount + 1;
        
        xt(1) = x(1)+a;
        xt(2) = x(2)+b;
        if xt(1)>N || xt(1)<1 || xt(2)>M || xt(2)<1
            continue;
        elseif Map2D(xt(1),xt(2))
            continue;
        end
        
        si = SpaceInd(x);
        sit = SpaceInd(xt);
        if si > sit
            arrowSgn = -1;
            tmpsi=si; si=sit; sit=tmpsi;
        else
            arrowSgn = 1;
        end
        if length(ComputedTransitions)>=si && length(ComputedTransitions{si})>=neiCount && ComputedTransitions{si}(neiCount)~=0
            dI = arrowSgn * ComputedTransitions{si}(neiCount);
        else
            dI = Delta_i(x, xt);
            ComputedTransitions{si}(neiCount) = arrowSgn * dI;
        end
        
        tryLval = ExistingLVals{x(1),x(2)}(xL) + dI;
        yi(end+1) = X2I(xt,tryLval);
    end
end

function ind = SpaceInd(x)
global N M
ind = (x(2)-1)*N + x(1);

% ======================================================================

function p = reconstruct_path(stop, start, TargetLVal)
global came_from

if length(stop) > 1
    stop = StopIndex;
    if nargin < 3
        TargetLVal = [];
    end
    currind = X2I(stop,TargetLVal);
    p = stop;
elseif length(stop)==1
    currind = stop;
    p = I2X(currind);
end

startind = X2I(start,0);

while came_from(currind)>0 %&& currind~=startind
    p(end+1,:) = I2X(came_from(currind));
    currind = came_from(currind);
end
